# -------------------------------------------------------------------------------
# Copyright IBM Corp. 2016
# 
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
# http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -------------------------------------------------------------------------------

from .display import GraphDisplay
from ..display import *

@PixiedustDisplay()
class GraphDisplayMeta(DisplayHandlerMeta):
    @addId
    def getMenuInfo(self,entity, dataHandler):
        clazz = entity.__class__.__name__
        if clazz == "GraphFrame":
            #Check that we have a longitude and latitude in the vertices dataframe
            fnames=[sf.name for sf in entity.vertices.schema.fields]
            ret = []
            if "longitude" in fnames and "latitude" in fnames:
                ret.append({"categoryId": "Graph", "title": "Graph Map", "icon":"fa-map-marker", "id":"graphMap"})
            ret.append({"categoryId": "Graph", "title": "Node-Link Graph", "icon":"fa-map-marker", "id":"nodeLinkGraph"})
            return ret
        
        return []
            
    def newDisplayHandler(self,options,entity):
        return GraphDisplay(options,entity)



# -------------------------------------------------------------------------------
# Copyright IBM Corp. 2016
# 
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
# http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -------------------------------------------------------------------------------
from pixiedust.utils.template import PixiedustTemplateEnvironment
from IPython.core.getipython import *
from IPython.display import display, HTML, Javascript
from pixiedust.utils.shellAccess import ShellAccess
from functools import reduce
import uuid
import json
import sys
import traceback
import pixiedust
from IPython.core.getipython import get_ipython
from collections import OrderedDict
from threading import Thread, Lock, Event
import time

myLogger = pixiedust.getLogger(__name__)
_env = PixiedustTemplateEnvironment()
progressMonitor = None
loadingProgressMonitor = False

def enableSparkJobProgressMonitor():
    global progressMonitor, loadingProgressMonitor
    if progressMonitor is None and not loadingProgressMonitor:
        loadingProgressMonitor = True
        def startSparkJobProgressMonitor():
            global progressMonitor
            progressMonitor = SparkJobProgressMonitor()
        t = Thread(target=startSparkJobProgressMonitor)
        t.daemon = True
        t.start()

class SparkJobProgressMonitorOutput(Thread):
    class Java:
        implements = ["com.ibm.pixiedust.PixiedustOutputListener"]
    
    def __init__(self):
        super(SparkJobProgressMonitorOutput,self).__init__()
        self.prefix = None
        self.lock = Lock()
        self.triggerEvent = Event()
        self.daemon = True
        self.progressData = OrderedDict()

    def getUpdaterId(self):
        return "updaterId{0}".format(self.prefix)

    def getProgressHTMLId(self):
        return "progress{0}".format(self.prefix)

    def run(self):
        while True:
            self.triggerEvent.wait()
            with self.lock:
                self.triggerEvent.clear()
                if bool(self.progressData):
                    progressData = self.progressData
                    self.progressData = OrderedDict()
                else:
                    progressData = OrderedDict()

            if bool(progressData):
                js = ""
                for data in progressData.values():
                    channel = data["channel"]
                    if channel=="jobStart":
                        js += _env.getTemplate("sparkJobProgressMonitor/addJobTab.js").render( 
                            prefix=self.prefix, data=data, overalNumTasks=reduce(lambda x,y:x+y["numTasks"], data["stageInfos"], 0) 
                        )
                    elif channel=="stageSubmitted":
                        js += _env.getTemplate("sparkJobProgressMonitor/updateStageStatus.js").render( 
                            prefix=self.prefix, stageId=data["stageInfo"]["stageId"], status="Submitted", host=None 
                        )
                    elif channel=="taskStart":
                        js += _env.getTemplate("sparkJobProgressMonitor/taskStart.js").render( prefix=self.prefix, data=data, increment = data["increment"] )
                        js += "\n"
                        js += _env.getTemplate("sparkJobProgressMonitor/updateStageStatus.js").render( 
                            prefix=self.prefix, stageId=data["stageId"], status="Running",
                            host="{0}({1})".format(data["taskInfo"]["executorId"],data["taskInfo"]["host"] )
                        )
                    elif channel=="stageCompleted":
                        js += _env.getTemplate("sparkJobProgressMonitor/updateStageStatus.js").render( 
                            prefix=self.prefix, stageId=data["stageInfo"]["stageId"], status="Completed", host=None 
                        )
                    elif channel=="jobEnd":
                        js += _env.getTemplate("sparkJobProgressMonitor/jobEnded.js").render( 
                            prefix=self.prefix, jobId=data["jobId"] 
                        )
                    js += "\n"

                display(Javascript(js))
            time.sleep(0.5)

    def display_with_id(self, obj, display_id, update=False):
        """Create a new display with an id"""
        ip = get_ipython()
        if hasattr(ip, "kernel"):
            data, md = ip.display_formatter.format(obj)
            content = {
                'data': data,
                'metadata': md,
                'transient': {'display_id': display_id},
            }
            msg_type = 'update_display_data' if update else 'display_data'
            ip.kernel.session.send(ip.kernel.iopub_socket, msg_type, content, parent=ip.parent_header)
        else:
            display(obj)

    def printOutput(self, s):
        print(s)

    def sendChannel(self, channel, data):
        self.printStuff(channel, data)

    def onRunCell(self):
        self.prefix = str(uuid.uuid4())[:8]
        #Create the place holder area for the progress monitor
        self.display_with_id( 
            HTML( _env.getTemplate("sparkJobProgressMonitor/pmLayout.html").render( prefix = self.prefix)),self.getProgressHTMLId() 
        )

    def printStuff(self,channel, s):
        try:
            data = json.loads(s)
            data["channel"] = channel
            data["increment"] = 1
            key = None
            if channel=="jobStart":
                key = "{0}-{1}".format(channel,data["jobId"])
            elif channel=="stageSubmitted":
                key = "{0}-{1}".format(channel,data["stageInfo"]["stageId"])
            elif channel=="taskStart":
                key = "{0}-{1}".format(channel,data["stageId"])
            elif channel=="stageCompleted":
                key = "{0}-{1}".format(channel,data["stageInfo"]["stageId"])
            elif channel=="jobEnd":
                key = "{0}-{1}".format(channel,data["jobId"])

            if key:
                with self.lock:
                    if key in self.progressData:
                        data["increment"] = self.progressData[key]["increment"] + 1
                    self.progressData[key] = data
                    self.triggerEvent.set()
        except:
            print("Unexpected error: {0} - {1} : {2}".format(channel, s, sys.exc_info()[0]))
            traceback.print_exc()

class SparkJobProgressMonitor(object):
    def __init__(self):
        self.monitorOutput = None
        self.addSparkListener()
        self.displayRuns={}
        self.newDisplayRun = False

    def onDisplayRun(self, contextId):
        if contextId is None or self.monitorOutput is None:
            self.newDisplayRun=True
            return

        cellContext = self.displayRuns.get( contextId )
        if cellContext and cellContext != self.monitorOutput.prefix:
            #switch the cell context if not a new display Run
            if self.newDisplayRun:
                self.displayRuns.pop( contextId, None )
            else:
                self.monitorOutput.prefix = cellContext
        elif cellContext is None:
            self.displayRuns[contextId] = self.monitorOutput.prefix

        if cellContext:
            display(Javascript(_env.getTemplate("sparkJobProgressMonitor/emptyTabs.js").render(prefix=cellContext)))

        self.newDisplayRun=False

    def addSparkListener(self):
        try:
            get_ipython().run_cell_magic(
                "scala",
                "cl=sparkProgressMonitor noSqlContext",
                _env.getTemplate("sparkJobProgressMonitor/addSparkListener.scala").render()
            )

            listener = get_ipython().user_ns.get("__pixiedustSparkListener")

            #access the listener object from the namespace
            if listener:
                self.monitorOutput = SparkJobProgressMonitorOutput()
                self.monitorOutput.start()
                #Add pre_run_cell event handler
                get_ipython().events.register('pre_run_cell',lambda: self.monitorOutput.onRunCell() )
                listener.setChannelListener( self.monitorOutput )
        except:
            myLogger.exception("Unexpected error while adding Spark Listener")
            raise



# -------------------------------------------------------------------------------
# Copyright IBM Corp. 2016
# 
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
# http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# Inherited from maven-artifact https://github.com/hamnis/maven-artifact
# -------------------------------------------------------------------------------

import os

class Package(object):
    def __init__(self, group_id, artifact_id, version):
        if not group_id:
            raise ValueError("group_id must be set")
        if not artifact_id:
            raise ValueError("artifact_id must be set")

        self.group_id = group_id
        self.artifact_id = artifact_id
        self.version = version
        self.uri = None

    def path(self, with_version=True):
        base = self.group_id.replace(".", "/") + "/" + self.artifact_id
        if with_version:
            return base + "/" + self.version
        else:
            return base

    def getUri(self, base, resolved_version=None):
        if self.uri:
            return self.uri
        if not resolved_version:
            resolved_version = self.version
        return base + "/" + self.path() + "/" + self.artifact_id + "-" + resolved_version + ".jar"

    def _generateFileName(self):
        s = self.artifact_id
        if self.version:
            s = s + "-" + self.version
        return s + ".jar"

    def getFilePath(self, filename=None):
        if  self.uri:
            return os.path.join(filename, self.uri.split("/")[-1])

        if not filename:
            filename = self._generateFileName()
        elif os.path.isdir(filename):
            filename = os.path.join(filename, self._generateFileName())
        return filename

    def __str__(self):
        if self.uri:
            return self.uri
        return "{0}:{1}:{2}".format(self.group_id, self.artifact_id, self.version)

    @staticmethod
    def clone(package, version=None):
        return Package(
            package.group_id,
            package.artifact_id,
            version if version is not None else package.version)

    @staticmethod
    def fromPackageIdentifier(package):
        #check if the user wants a direct download
        if package.startswith("http://") or package.startswith("https://") or package.startswith("file://"):
            retPackage = Package( "direct.download", package, "1.0" )
            retPackage.uri = package
            return retPackage

        parts = package.split(":")
        if len(parts) >= 3:
            g = parts[0]
            a = parts[1]
            v = parts[len(parts) - 1]
            return Package(g, a, v)
        return None



# -------------------------------------------------------------------------------
# Copyright IBM Corp. 2016
# 
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
# http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -------------------------------------------------------------------------------

from pixiedust.display.chart.renderers import PixiedustRenderer
from pixiedust.display.chart.renderers.baseChartDisplay import commonChartOptions
from .matplotlibBaseDisplay import MatplotlibBaseDisplay
from pixiedust.utils import Logger
import matplotlib.pyplot as plt
import mpld3
import numpy as np

@PixiedustRenderer(id="pieChart")
@Logger()
class PieChartDisplay(MatplotlibBaseDisplay):

    def supportsKeyFieldLabels(self, handlerId):
        return False
    
    def getPreferredDefaultValueFieldCount(self, handlerId):
        return 1

    def getDefaultAggregation(self, handlerId):
        return "COUNT"

    def canStretch(self):
        return False

    # override the default keys displayed when the chart first renders
    def getDefaultKeyFields(self, handlerId, aggregation):
        return self.sampleColumn((aggregation != "COUNT"))

    # override the default values displayed when the chart first renders
    def getDefaultValueFields(self, handlerId, aggregation):
        return self.getDefaultKeyFields(handlerId, aggregation)

    def setChartGrid(self, fig, ax):
        pass

    @commonChartOptions
    def getChartOptions(self):
        return [
            {
                'name': 'legend',
                'description': 'Show legend',
                'metadata': {
                    'type': 'checkbox',
                    'default': "false"
                }
            }
        ]

    def getNumFigures(self):
        return len(self.getValueFields())

    """
        We want height equals to size to make a perfect pie
    """
    def getHeightWidthRatio(self):
        return 1

    """
        return chart width scale factor
    """
    def getWidthScaleFactor(self):
        return 0.6

    def getSubplotHSpace(self):
        return None

    def isStretchingOn(self):
        return True

    def matplotlibRender(self, fig, ax):
        numRows = len(self.getWorkingPandasDataFrame().index)
        if numRows > 20:
            self.addMessage("Too many data points to plot. Dropping {0} rows to make the chart more presentable.".format(numRows-20))
        if not isinstance(ax, (list,np.ndarray)):
            ax=np.array([ax])
        keyFields = self.getKeyFields()
        valueFields = self.getValueFields()
        for i,valueField in enumerate(valueFields):
            pdf = self.getWorkingPandasDataFrame().sort_values(valueField).head(20)
            labels=[ "-".join(map(str, a)) for a in pdf[keyFields].values.tolist() ]
            pdf.plot(
                kind="pie", y = valueField, ax=ax.item(i), labels=labels, 
                autopct='%1.0f%%', subplots=False, legend = self.showLegend()
            )


# -------------------------------------------------------------------------------
# Copyright IBM Corp. 2017
# 
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
# http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -------------------------------------------------------------------------------

from ..display import ChartDisplay
from abc import abstractmethod, ABCMeta
import traceback
import pixiedust
from collections import OrderedDict
from six import PY2, with_metaclass,iteritems
from pixiedust.display import addDisplayRunListener
from pixiedust.display.chart.renderers import PixiedustRenderer
from pixiedust.utils import cache,Logger
from pixiedust.utils.shellAccess import ShellAccess
from .commonOptions import commonOptions as co
import pandas as pd
import time
import inspect
import re

class ShowChartOptionDialog(Exception):
    pass

class WorkingDataCache(with_metaclass( 
        type("",(type,),{
            "workingDataCache":{},
            "__getitem__":lambda cls, key: cls.workingDataCache.get(key),
            "__setitem__":lambda cls, key,val: cls.workingDataCache.update({key:val}),
            "__getattr__":lambda cls, key: cls.workingDataCache.get(key),
            "__setattr__":lambda cls, key, val: cls.workingDataCache.update({key:val})
        }), object
    )):

    myLogger = pixiedust.getLogger(__name__)

    @staticmethod
    def onNewDisplayRun(entity, options):
        if "cell_id" in options and "showchrome" in options:
            #User is doing a new display run from the cell, clear the cache as we don't know if the entity has changed
            WorkingDataCache.removeEntry(options["cell_id"])

    @staticmethod
    def removeEntry(key):
        WorkingDataCache.workingDataCache.pop(key, None)

    @staticmethod
    def putInCache(options, data, constraints):
        if "cell_id" not in options:
            return
        constraints.pop("self",None)
        WorkingDataCache[options["cell_id"]] = {
            "data": data,
            "constraints": constraints
        }

    @staticmethod
    def getFromCache(options, constraints ):
        if "cell_id" not in options:
            return None
        constraints.pop("self",None)
        cellId = options["cell_id"]
        value = WorkingDataCache[cellId]
        if value:
            WorkingDataCache.myLogger.debug("Found cache data for {}. Validating integrity...".format(cellId))
            for item in list(value["constraints"].items()):
                if item[0] not in constraints or item[1] != constraints[item[0]]:
                    WorkingDataCache.myLogger.debug("Cache data not validated for key {0}. Expected Value is {1}. Got {2}. Destroying it!...".format(item[0], item[1], constraints[item[0]]))
                    WorkingDataCache.removeEntry(cellId)
                    return None
            WorkingDataCache.myLogger.debug("Cache data validated for {}. Using it!...".format(cellId))
            return value["data"]
        WorkingDataCache.myLogger.debug("No Cache Entry found for {}".format(cellId))

#add a display Run Listener 
addDisplayRunListener( lambda entity, options: WorkingDataCache.onNewDisplayRun(entity, options) )

#common Chart Options injection decorator
def commonChartOptions(func):
    def wrapper(cls, *args, **kwargs):
        commonOptions = []
        if hasattr(cls, "handlerId") and hasattr(cls, 'commonOptions'):
            handler = cls.commonOptions.get(cls.handlerId, [])
            if callable(handler):
                handler = handler(cls)
            commonOptions += handler
            fctSet = set()
            for cl in reversed(inspect.getmro(cls.__class__)):
                if hasattr(cl, func.__name__):
                    f = getattr(cl, func.__name__)
                    if f not in fctSet:
                        fctSet.add(f)
                        if hasattr(f, "func"):
                            f = f.func
                        opts = f(cls, *args, **kwargs)
                        commonOptions += opts
            return commonOptions
        return commonOptions + func(cls, *args, **kwargs )
    wrapper.func = func
    return wrapper

@Logger()
class BaseChartDisplay(with_metaclass(ABCMeta, ChartDisplay)):

    commonOptions = co

    def __init__(self, options, entity, dataHandler=None):
        super(BaseChartDisplay,self).__init__(options,entity,dataHandler)
        #note: since this class can be subclassed from other module, we need to mark the correct resource module with resModule so there is no mixup
        self.extraTemplateArgs["resModule"]=BaseChartDisplay.__module__
        self.messages = []

    """
        Subclass can override: return an array of option metadata
    """
    @commonChartOptions
    def getChartOptions(self):
        return []

    def validateOptions(self):
        #validate options
        chartOptions = self.getChartOptions()   
        self.debug("chartOptions {}".format(chartOptions)) 
        ord = OrderedDict([(o["name"],o["validate"]) for o in chartOptions if "validate" in o and "name" in o])
        remKeys = []
        for key,value in iteritems(self.options):
            if key in ord:
                values = value.split(",")
                self.debug("values: {0}".format(values))
                for v in values:
                    self.debug("Calling with {0}".format(v))
                    passed, message = ord.get(key)(v)
                    if not passed:
                        self.addMessage("Filtered option {0} with value {1}. Reason {2}".format(key, value, message))
                        remKeys.append(key)
                        break

        for key in remKeys:
            del self.options[key]

    def getExtraFields(self):
        return []

    def addMessage(self, message):
        self.messages.append(message)

    @cache(fieldName="workingPandasDataFrame")
    def getWorkingPandasDataFrame(self):
        xFields = self.getKeyFields()
        yFields = self.getValueFields()
        extraFields = self.getExtraFields()
        aggregation = self.getAggregation()
        maxRows = self.getMaxRows()
        #remember the constraints for this cache, they are the list of variables
        constraints = locals()

        workingDF = WorkingDataCache.getFromCache(self.options, constraints )
        if workingDF is None:
            workingDF = self.dataHandler.getWorkingPandasDataFrame(xFields, yFields, extraFields = extraFields, aggregation=aggregation, maxRows = maxRows )
            WorkingDataCache.putInCache(self.options, workingDF, constraints)
        
        if self.options.get("debug", None):
            self.debug("getWorkingPandasDataFrame returns: {0}".format(workingDF) )
            ShellAccess["workingPDF"] = workingDF    
        return workingDF

    def getWorkingDataSlice1( self, col, sort = False ):
        colData = self.getWorkingPandasDataFrame()[col].values.tolist()
        if sort:
            return sorted(colData)
        else:
            return colData

    def getWorkingDataSlice( self, col1, col2, sort = False ):
        col1Data = self.getWorkingPandasDataFrame()[col1].values.tolist()
        col2Data = self.getWorkingPandasDataFrame()[col2].values.tolist()
        if sort:
            return zip(*sorted(zip(col1Data, col2Data )))
        else:
            return [col1Data, col2Data]

    @cache(fieldName="maxRows")
    def getMaxRows(self):
        return int(self.options.get("rowCount","100"))

    # def getPandasDataFrame(self):
    #     valueFieldValues = self.getValueFieldValueLists()
    #     valueFields = self.getValueFields()
    #     return pd.DataFrame([list(a) for a in zip( valueFieldValues[0], valueFieldValues[1]) ], columns=[valueFields[0], valueFields[1]])

    #helper method
    def _getField(self, fieldName):
        if not hasattr(self, fieldName):
            return None
        return getattr(self, fieldName)

    def getChartContext(self, handlerId):
        return None

    def getChartErrorDialogBody(self, handlerId, dialogTemplate, dialogOptions):
        return self.renderTemplate(dialogTemplate, **dialogOptions)

    @abstractmethod
    def doRenderChart(self):
        pass

    def isMap(self, handlerId):
        return False

    def supportsKeyFields(self, handlerId):
        return True

    def supportsKeyFieldLabels(self, handlerId):
        return True

    def supportsLegend(self, handlerId):
        return True

    def supportsAggregation(self, handlerId):
        return True

    def getDefaultAggregation(self, handlerId):
        return "SUM"

    def getPreferredDefaultKeyFieldCount(self, handlerId):
        return 1

    @cache(fieldName="keyFields")
    def getKeyFields(self):
        """ Get the dataframe field names from metadata (usually specified by the user 
        as key fields in the chart option dialog)

        Args: 
            self (class): class that extends BaseChartDisplay

        Returns: 
            List of strings: dataframe field names

        Raises:
            Calls ShowChartOptionDialog() if array is empty
        """
        fieldNames = self.getFieldNames() # get all field names in data format-independent way
        if self.supportsKeyFields(self.handlerId) == False:
            return []
        keyFields = []
        keyFieldStr = self.options.get("keyFields")
        if keyFieldStr is not None:
            keyFields = keyFieldStr.split(",")
            keyFields = [val for val in keyFields if val in fieldNames]
        if len(keyFields) == 0:
            raise ShowChartOptionDialog()
        else:
            return keyFields

    @cache(fieldName="keyFieldLabels")
    def getKeyFieldLabels(self):
        k = self.getKeyFields()
        if (len(k) == 0):
            return []
        return self.getWorkingPandasDataFrame().groupby(k).groups.keys()

    def getPreferredDefaultValueFieldCount(self, handlerId):
        return 2

    @cache(fieldName="valueFields")   
    def getValueFields(self):
        fieldNames = self.getFieldNames()
        aggregation = self.getAggregation()
        valueFields = []
        valueFieldStr = self.options.get("valueFields")
        if valueFieldStr is not None:
            valueFields = valueFieldStr.split(",")
            valueFields = [val for val in valueFields if val in fieldNames]
        numericValueFields = []
        for valueField in valueFields:
            if self.dataHandler.isNumericField(valueField) or aggregation == "COUNT":
                numericValueFields.append(valueField)
        if len(numericValueFields) == 0:
            raise ShowChartOptionDialog()
        else:
            return numericValueFields
    
    def canRenderChart(self):
        aggregation = self.getAggregation()
        if (aggregation == "COUNT"):
            return (True, None)
        else:
            for field in self.dataHandler.schema.fields:
                if self.dataHandler.isNumericType(field):
                    return (True, None)
            return (False, "At least one numerical column required.")

    def getDialogInfo(self, handlerId):
        context = self.getChartContext(handlerId)
        dialogOptions = { "fieldNames":self.getFieldNames(True),\
            "fieldNamesAndTypes":self.getFieldNamesAndTypes(True, True),\
            "keyFieldsSupported":self.supportsKeyFields(handlerId),\
            "legendSupported":self.supportsLegend(handlerId),\
            "aggregationSupported":self.supportsAggregation(handlerId),\
            "aggregationOptions":["SUM","AVG","MIN","MAX","COUNT"]\
        }
        if (context is not None):
            dialogTemplate = context[0]
            dialogOptions.update(context[1])
        else:
            dialogTemplate = BaseChartDisplay.__module__ + ":baseChartOptionsDialogBody.html"

        return (dialogTemplate, dialogOptions)

    def getRendererList(self):
        return PixiedustRenderer.getRendererList(self.options, self.entity)

    @cache(fieldName="aggregation")
    def getAggregation(self):
        if not self.supportsAggregation(self.handlerId):
            return None
        # get aggregation value (set to default if it doesn't exist)
        aggregation = self.options.get("aggregation")
        if aggregation is None:
            aggregation = self.getDefaultAggregation(self.handlerId)
            self.options["aggregation"] = aggregation
        return aggregation

    @cache(fieldName="fieldNames")
    def getFieldNames(self, expandNested=True):
        return self.dataHandler.getFieldNames(expandNested)

    @cache(fieldName="fieldNamesAndTypes")
    def getFieldNamesAndTypes(self, expandNested=True, sorted=False):
        fieldNames = self.getFieldNames(True)
        fieldNamesAndTypes = []
        for fieldName in fieldNames:
            fieldType = "unknown/unsupported"
            if self.dataHandler.isNumericField(fieldName):
                fieldType = "numeric"
            if self.dataHandler.isStringField(fieldName):
                fieldType = "string"
            fieldNamesAndTypes.append((fieldName, fieldType))
        if sorted:
            fieldNamesAndTypes.sort(key=lambda x: x[0])
        return fieldNamesAndTypes

    def doRender(self, handlerId):
        self.handlerId = handlerId
        optionsTitle = self.camelCaseSplit(handlerId, True) + " Options"
        if self.options.get("debug", None):
            self.logStuff()

        # field names
        fieldNames = self.getFieldNames(True)
        (dialogTemplate, dialogOptions) = self.getDialogInfo(handlerId)

        # go
        try:
            self.validateOptions()
            keyFields = self.getKeyFields()
            valueFields = self.getValueFields()
        except ShowChartOptionDialog:
            self.dialogBody = self.renderTemplate(dialogTemplate, **dialogOptions)
            self._addJavascriptTemplate("chartOptions.dialog", optionsDialogBody=self.dialogBody, optionsTitle=optionsTitle, inScript=True)
            return
        
        # render
        try:
            self.dialogBody = self.renderTemplate(dialogTemplate, **dialogOptions)

            # validate if we can render
            canRender, errorMessage = self.canRenderChart()
            if canRender == False:
                raise Exception(errorMessage)

            # set the keyFields and valueFields options if this is the first time
            # do this after call to canRenderChart as some charts may need to know that they have not been set
            setKeyFields = self.options.get("keyFields") is None
            setValueFields = self.options.get("valueFields") is None
            if setKeyFields and len(keyFields) > 0:
                self.options["keyFields"] = ",".join(keyFields)
            if setValueFields and len(valueFields) > 0:
                self.options["valueFields"] = ",".join(valueFields)        

            chartFigure = self.doRenderChart()
            if self.options.get("debugFigure", None):
                self.debug(chartFigure)
            if self.options.get("nostore_figureOnly", None):
                self._addHTML(chartFigure)
            else:
                self._addHTMLTemplate("renderer.html", chartFigure=chartFigure, optionsDialogBody=self.dialogBody, optionsTitle=optionsTitle)
        except Exception as e:
            self.exception("Unexpected error while trying to render BaseChartDisplay")
            errorHTML = """
                <div style="min-height: 50px;">
                    <div style="color: red;position: absolute;bottom: 0;left: 0;">{0}</div>
                </div>
            """.format(str(e))
            if self.options.get("nostore_figureOnly", None):
                self._addHTML(errorHTML)
            else:
                self._addHTMLTemplate("renderer.html", chartFigure=errorHTML, optionsDialogBody=self.dialogBody, optionsTitle=optionsTitle)

    def logStuff(self):
        try:
            self.debug("Key Fields: {0}".format(self.getKeyFields()) )
            ShellAccess['keyFields'] = self.getKeyFields()
            self.debug("Key Fields Labels: {0}".format(self.getKeyFieldLabels()))
            self.debug("Value Fields: {0}".format(self.getValueFields()))
            ShellAccess['valueFields'] = self.getValueFields()
        except:
            pass

    def camelCaseSplit(self, camelCaseStr, titleCase=False):
        matches = re.finditer('.+?(?:(?<=[a-z])(?=[A-Z])|(?<=[A-Z])(?=[A-Z][a-z])|$)', camelCaseStr)
        split = " ".join(m.group(0) for m in matches)
        if titleCase:
            return split.title()
        else:
            return split



# -------------------------------------------------------------------------------
# Copyright IBM Corp. 2016
# 
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
# http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -------------------------------------------------------------------------------

import pixiedust
from pixiedust.utils.userPreferences import *

myLogger = pixiedust.getLogger(__name__)

__all__ = ['PixiedustRenderer']

"""global dictionary used to register the renderers for each display id"""
_renderers = {}

"""
PixieDust renderer decorator. Recommended pattern is to decorate a base class with the rendererId (e.g. matplotlib)
and have all renderer subclasses provide the id
"""
class PixiedustRenderer(object):
    def __init__(self, id=None, rendererId=None):
        self.id = id
        self.rendererId = rendererId

    def __call__(self, cls, *args, **kwargs):
        if not hasattr(cls, "_id") or self.id is not None:
            cls._id = self.id
        if not hasattr(cls, "_rendererId") or self.rendererId is not None:
            cls._rendererId = self.rendererId

        if cls._id is not None and cls._rendererId is not None:
            if _renderers.get(cls._id) is None:
                _renderers[cls._id] = [cls]
            else:
                _renderers[cls._id].append(cls)
        return cls

    @staticmethod
    def getRenderer(options, entity):
        handlerId = options.get("handlerId")
        rendererId = options.get('rendererId', None )
        if rendererId is not None:
            setUserPreference(handlerId, rendererId)
        else:
            rendererId = getUserPreference( handlerId, 'matplotlib')
        renderers = _renderers.get(handlerId)
        if renderers is None or len(renderers)==0:
            myLogger.debug("Couldn't find a renderer for {0} in {1}".format(handlerId, _renderers))
            raise Exception("No renderer available for {0}".format(handlerId) )
        #Lookup the renderer, default to first one if not found
        for renderer in renderers:
            if renderer._rendererId == rendererId:
                myLogger.debug("Found renderer: {0} - {1}".format(renderer._rendererId, renderer._id))
                return renderer(options, entity)
        myLogger.debug("Defaulting to first renderer {0} - {1}".format(renderers[0]._rendererId, renderers[0]._id))
        return renderers[0](options, entity)

    @staticmethod
    def getRendererList(options, entity):
        handlerId = options.get("handlerId")
        if handlerId is None:
            return []
        return _renderers.get(handlerId)


# -------------------------------------------------------------------------------
# Copyright IBM Corp. 2016
# 
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
# http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -------------------------------------------------------------------------------

from .display import TableDisplay
from ..display import *
import pixiedust.utils.dataFrameMisc as dataFrameMisc

@PixiedustDisplay(isDefault=True)
class TableDisplayMeta(DisplayHandlerMeta):
    @addId
    def getMenuInfo(self,entity, dataHandler):
        if dataFrameMisc.isPySparkDataFrame(entity) or dataFrameMisc.isPandasDataFrame(entity):
            return [
                {"categoryId": "Table", "title": "DataFrame Table", "icon": "fa-table", "id": "dataframe"}
            ]
        elif dataFrameMisc.fqName(entity) == "graphframes.graphframe.GraphFrame":
            return [
                {"categoryId": "Table", "title": "Graph Vertices", "icon": "fa-location-arrow", "id":"vertices"},
                {"categoryId": "Table", "title": "Graph Edges", "icon": "fa-link", "id":"edges"}
            ]
        else:
            return []
    def newDisplayHandler(self,options,entity):
        return TableDisplay(options,entity)


# -------------------------------------------------------------------------------
# Copyright IBM Corp. 2016
# 
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
# http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -------------------------------------------------------------------------------
from six import iteritems
from pyspark.sql.types import *
import pixiedust
from pixiedust.utils.shellAccess import ShellAccess
from pixiedust.utils.template import PixiedustTemplateEnvironment
from pixiedust.utils.environment import Environment,scalaGateway
import uuid
import tempfile
from collections import OrderedDict
from IPython.display import display, HTML, Javascript
try:
    from urllib.request import Request, urlopen, URLError, HTTPError
except ImportError:
    from urllib2 import Request, urlopen, URLError, HTTPError

dataDefs = OrderedDict([
    ("1", {
        "displayName": "Car performance data", 
        "url": "https://github.com/ibm-cds-labs/open-data/raw/master/cars/cars.csv",
        "topic": "transportation",
        "publisher": "IBM",
        "schema2": [('mpg','int'),('cylinders','int'),('engine','double'),('horsepower','int'),('weight','int'),
            ('acceleration','double'),('year','int'),('origin','string'),('name','string')]
    }),
    ("2", {
        "displayName": "Sample retail sales transactions, January 2009", 
        "url": "https://raw.githubusercontent.com/ibm-cds-labs/open-data/master/salesjan2009/salesjan2009.csv",
        "topic": "Economy & Business",
        "publisher": "IBM Cloud Data Services"
    }),
    ("3", {
        "displayName": "Total population by country", 
        "url": "https://apsportal.ibm.com/exchange-api/v1/entries/889ca053a19986a4445839358a91963e/data?accessKey=657b130d504ab539947e51b50f0e338e",
        "topic": "Society",
        "publisher": "IBM Cloud Data Services"
    }),
    ("4", {
        "displayName": "GoSales Transactions for Naive Bayes Model", 
        "url": "https://apsportal.ibm.com/exchange-api/v1/entries/8044492073eb964f46597b4be06ff5ea/data?accessKey=bec2ed69d9c84bed53826348cdc5690b",
        "topic": "Leisure",
        "publisher": "IBM"
    }),
    ("5", {
        "displayName": "Election results by County", 
        "url": "https://openobjectstore.mybluemix.net/Election/county_election_results.csv",
        "topic": "Society",
        "publisher": "IBM"
    }),
    ("6", {
        "displayName": "Million dollar home sales in NE Mass late 2016", 
        "url": "https://openobjectstore.mybluemix.net/misc/milliondollarhomes.csv",
        "topic": "Economy & Business",
        "publisher": "Redfin.com"
    }),
    ("7", {
        "displayName": "Boston Crime data, 2-week sample", 
        "url": "https://raw.githubusercontent.com/ibm-cds-labs/open-data/master/crime/boston_crime_sample.csv",
        "topic": "Society",
        "publisher": "City of Boston"
    })
])

@scalaGateway
def sampleData(dataId=None):
    global dataDefs
    return SampleData(dataDefs).sampleData(dataId)

class SampleData(object):
    env = PixiedustTemplateEnvironment()
    def __init__(self, dataDefs):
        self.dataDefs = dataDefs

    def sampleData(self, dataId = None):
        if dataId is None:
            self.printSampleDataList()
        elif str(dataId) in dataDefs:
            return self.loadSparkDataFrameFromSampleData(dataDefs[str(dataId)])
        elif "https://" in str(dataId) or "http://" in str(dataId):
            return self.loadSparkDataFrameFromUrl(str(dataId))
        else:
            print("Unknown sample data identifier. Please choose an id from the list below")
            self.printSampleDataList()

    def printSampleDataList(self):
        display( HTML( self.env.getTemplate("sampleData.html").render( dataDefs = iteritems(self.dataDefs) ) ))
        #for key, val in iteritems(self.dataDefs):
        #    print("{0}: {1}".format(key, val["displayName"]))

    def dataLoader(self, path, schema=None):
        #TODO: if in Spark 2.0 or higher, use new API to load CSV
        load = ShellAccess["sqlContext"].read.format('com.databricks.spark.csv')
        if schema is not None:
            def getType(t):
                if t == 'int':
                    return IntegerType()
                elif t == 'double':
                    return DoubleType()
                else:
                    return StringType()
            return load.options(header='true', mode="DROPMALFORMED").load(path, schema=StructType([StructField(item[0], getType(item[1]), True) for item in schema]))
        else:
            return load.options(header='true', mode="DROPMALFORMED", inferschema='true').load(path)

    def loadSparkDataFrameFromSampleData(self, dataDef):
        return Downloader(dataDef).download(self.dataLoader)

    def loadSparkDataFrameFromUrl(self, dataUrl):
        i = dataUrl.rfind('/')
        dataName = dataUrl[(i+1):]
        dataDef = {
            "displayName": dataUrl,
            "url": dataUrl
        }
        return Downloader(dataDef).download(self.dataLoader)


class Downloader(object):
    def __init__(self, dataDef):
        self.dataDef = dataDef
        self.headers = {"User-Agent": "PixieDust Sample Data Downloader/1.0"}
        self.prefix = str(uuid.uuid4())[:8]
    
    def download(self, dataLoader):
        displayName = self.dataDef["displayName"]
        if "path" in self.dataDef:
            path = self.dataDef["path"]
        else:
            url = self.dataDef["url"]
            req = Request(url, None, self.headers)
            print("Downloading '{0}' from {1}".format(displayName, url))
            with tempfile.NamedTemporaryFile(delete=False) as f:
                self.write(urlopen(req), f)
                path = f.name
                self.dataDef["path"] = path = f.name
        if path:
            try:
                print("Creating pySpark DataFrame for '{0}'. Please wait...".format(displayName))
                return dataLoader(path, self.dataDef.get("schema", None))
            finally:
                print("Successfully created pySpark DataFrame for '{0}'".format(displayName))
            
    def report(self, bytes_so_far, chunk_size, total_size):
        if bytes_so_far == 0:
            display( HTML( """
                <div>
                    <span id="pm_label{0}">Starting download...</span>
                    <progress id="pm_progress{0}" max="100" value="0" style="width:200px"></progress>
                </div>""".format(self.prefix)
                )
            )
        else:
            percent = float(bytes_so_far) / total_size
            percent = round(percent*100, 2)
            display(
                Javascript("""
                    $("#pm_label{prefix}").text("{label}");
                    $("#pm_progress{prefix}").attr("value", {percent});
                """.format(prefix=self.prefix, label="Downloaded {0} of {1} bytes".format(bytes_so_far, total_size), percent=percent))
            )

    def write(self, response, file, chunk_size=8192):
        total_size = response.headers['Content-Length'].strip() if 'Content-Length' in response.headers else 100
        total_size = int(total_size)
        bytes_so_far = 0

        self.report(bytes_so_far, chunk_size, total_size)

        while 1:
            chunk = response.read(chunk_size)
            bytes_so_far += len(chunk)
            if not chunk:
                break
            file.write(chunk)             
            total_size = bytes_so_far if bytes_so_far > total_size else total_size
            self.report(bytes_so_far, chunk_size, total_size)

        return bytes_so_far


# -------------------------------------------------------------------------------
# Copyright IBM Corp. 2016
# 
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
# http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -------------------------------------------------------------------------------

from ..display.display import DisplayHandlerMeta,PixiedustDisplay,addId

from .stashCloudant import StashCloudantHandler
from .stashSwift import StashSwiftHandler
import pixiedust.utils.dataFrameMisc as dataFrameMisc

@PixiedustDisplay(system=True)
class StashMeta(DisplayHandlerMeta):
    @addId
    def getMenuInfo(self,entity, dataHandler):
        if dataFrameMisc.isPySparkDataFrame(entity):
            return [
                {"categoryId": "Download", "title": "Stash to Cloudant", "icon": "fa-cloud", "id": "stashCloudant"}
            ]
            #{"categoryId": "Download", "title": "Stash to Object Storage", "icon": "fa-suitcase", "id": "stashSwift"}
        else:
            return []
    def newDisplayHandler(self,options,entity):
        handlerId = options.get("handlerId")
        if handlerId=="stashCloudant":
            return StashCloudantHandler(options, entity)
        else:
            return StashSwiftHandler(options,entity)



# -------------------------------------------------------------------------------
# Copyright IBM Corp. 2017
# 
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
# http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -------------------------------------------------------------------------------
from six import with_metaclass
import math
import matplotlib as mpl
from matplotlib import cm

'''
Static class that provides consistent color palette for all charts
'''
class Colors(with_metaclass( 
        type("",(type,),{
            "colormap":cm.jet,
            "__getitem__":lambda cls, key: cls.colormap(key),
            "__getattr__":lambda cls, key: cls.colormap(key)
        }), object
    )):

    @staticmethod
    def hexRGB(X):
        tupleColor = Colors.colormap(X)
        return str(mpl.colors.rgb2hex( tupleColor ).upper())

    #TODO: provide user configured colormap palette
    pass



# -------------------------------------------------------------------------------
# Copyright IBM Corp. 2017
# 
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
# http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -------------------------------------------------------------------------------
from six import with_metaclass
from . import *
import os
import re
import subprocess

class Environment(with_metaclass( 
        type("",(type,),{
            "__getattr__":lambda cls, key: getattr(cls.env, key)
        }), object
    )):

    class _Environment(object):
        @property
        @cache(fieldName="_scalaVersion")
        def scalaVersion(self):
            scala = "{0}{1}bin{1}scala".format(self.scalaHome, os.sep)
            try:
                scala_out = subprocess.check_output([scala, "-version"], stderr=subprocess.STDOUT).decode("utf-8")
            except subprocess.CalledProcessError as cpe:
                scala_out = cpe.output
            match = re.search(".*version[^0-9]*([0-9]*[^.])\.([0-9]*[^.])\.([0-9]*[^.]).*", str(scala_out))
            if match and len(match.groups()) > 2:
                return int(match.group(1)), int(match.group(2))
            else:
                raise EnvironmentError("Unable to compute Scala Version")

        @property
        @cache(fieldName="_scalaHome")
        def scalaHome(self):
            return os.environ.get("SCALA_HOME")

    env = _Environment()

"""
Decorator for functions that can be called from Scala Notebook by adding the fromScala keyword argument to the decorated function
"""
def scalaGateway(func):
    def wrapper(*args,**kwargs):
        fromScala = False
        if "fromScala" in kwargs:
            kwargs.pop("fromScala")
            fromScala = True
        retValue = func(*args, **kwargs)
        if fromScala and retValue is not None:
            from pixiedust.utils.javaBridge import JavaWrapper
            from pixiedust.utils.scalaBridge import ConverterRegistry
            import uuid
            id = str(uuid.uuid4())[:8]
            JavaWrapper("com.ibm.pixiedust.Pixiedust").addEntity( id, ConverterRegistry.toJava( retValue )[0] )
            return id
        return retValue
    return wrapper



# -------------------------------------------------------------------------------
# Copyright IBM Corp. 2016
# 
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
# http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -------------------------------------------------------------------------------
import mpld3.plugins as plugins
from mpld3 import utils

class ElementInfoPlugin(plugins.PluginBase):
    
    JAVASCRIPT = """
    mpld3.register_plugin("elementInfo", ElementInfoPlugin);
    ElementInfoPlugin.prototype = Object.create(mpld3.Plugin.prototype);
    ElementInfoPlugin.prototype.constructor = ElementInfoPlugin;
    ElementInfoPlugin.prototype.requiredProps = ["id","info"];
    function ElementInfoPlugin(fig, props){
        mpld3.Plugin.call(this, fig, props);
    };
    
    ElementInfoPlugin.prototype.draw = function(){
        var obj = mpld3.get_element(this.props.id);
        var info = this.props.info;
        if (obj ){
            obj.elements()
                .each(function(d,i){
                    d3.select(this).attr("color",this.style.fill)
                })
                .on("mouseover", function(d) {
                    d3.select(this).style("fill", "#504554").style("stroke","white")
                })                  
                .on("mouseout", function(d) {
                    d3.select(this).style("fill", d3.select(this).attr("color")).style("stroke","black")
                })
                .append("svg:title")
                .text(function(d) { return info; })
        }else{
            console.log("Unable to find element with id",this.props.id);
        }
    }
    """
    def __init__(self, element,info):
        self.dict_ = {"type": "elementInfo",
                      "id": utils.get_id(element),
                      "info":info}


# -------------------------------------------------------------------------------
# Copyright IBM Corp. 2016
# 
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
# http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -------------------------------------------------------------------------------

import sqlite3
import os

import json
import sys
import time
from pixiedust.utils.constants import PIXIEDUST_REPO_URL
from pkg_resources import get_distribution
from re import search
from requests import post
from os import environ as env
from pixiedust.utils.printEx import *

from . import pdLogging
logger = pdLogging.getPixiedustLogger()
getLogger = pdLogging.getLogger

myLogger = getLogger(__name__)

__all__ = ['Storage']

SQLITE_DB_NAME = 'pixiedust.db'
SQLITE_DB_NAME_PATH = os.environ.get("PIXIEDUST_HOME", os.path.expanduser('~')) + "/" + SQLITE_DB_NAME

if not os.path.exists(os.path.dirname(SQLITE_DB_NAME_PATH)):
    os.makedirs(os.path.dirname(SQLITE_DB_NAME_PATH))

#global connection
_conn = None

def _initStorage():
    def copyRename(oldName, newName):
        if os.path.isfile(oldName) and not os.path.isfile(newName):
            from shutil import copyfile
            copyfile(oldName, newName)
            os.rename(oldName, oldName + ".migrated")
            print("INFO: migrated pixiedust database")

    #if db already exist with old name, rename it now
    copyRename('spark.db', SQLITE_DB_NAME)
    copyRename(SQLITE_DB_NAME, SQLITE_DB_NAME_PATH)

    global _conn
    if not _conn:
        def _row_dict_factory(cursor,row):
            res={}
            for i,col in enumerate(cursor.description):
                res[col[0]]=row[i]
            return res
        _conn = sqlite3.connect(SQLITE_DB_NAME_PATH)
        _conn.row_factory=_row_dict_factory 
        print("Pixiedust database opened successfully")

    _trackDeployment()

"""
Encapsule access to data from the pixiedust database
including storage lifecycle e.g. schema definition and creation, cleanup, etc...
"""
class Storage(object):
    def __init__(self):
        pass

    def _initTable(self, tableName, schemaDef):
        cursor=_conn.execute("""
            SELECT * FROM sqlite_master WHERE name ='{0}' and type='table';
        """.format(tableName))
        if cursor.fetchone() is None:
            _conn.execute('''CREATE TABLE {0} ({1});'''.format(tableName, schemaDef))
            print("Table {0} created successfully".format(tableName))
        '''
        else:
            print("Deleting table")
            _conn.execute("""DROP TABLE {0}""".format(tableName))
        '''
        cursor.close()

    def fetchOne(self, sqlQuery, mapper=None):
        cursor=None
        try:
            cursor=_conn.execute(sqlQuery)
            row = cursor.fetchone()
            return row if row is None or mapper is None else mapper(row)
        finally:
            if cursor is not None:
                cursor.close()

    def fetchMany(self, sqlQuery, mapper=None):
        cursor=None
        try:
            cursor=_conn.execute(sqlQuery)
            results = cursor.fetchmany()
            ret = []
            while results:
                for row in results:
                    ret.append( row if mapper is None else mapper(row) )
                results = cursor.fetchmany()
            return ret
        finally:
            if cursor is not None:
                cursor.close()

    def execute(self, sqlQuery, handler):
        cursor=None
        try:
            cursor=_conn.execute(sqlQuery)
            results = cursor.fetchmany()
            while results:
                for row in results:
                    handler(row)
                results = cursor.fetchmany()
        finally:
            if cursor is not None:
                cursor.close()

    """Return the number of rows deleted"""
    def delete(self, sqlQuery):
        cursor = None
        try:
            cursor=_conn.execute(sqlQuery)
            _conn.commit()
            return _conn.total_changes
        finally:
            if cursor is not None:
                cursor.close()

    def insert(self, sqlQuery):
        _conn.execute(sqlQuery)
        _conn.commit()

    def update(self, sqlQuery):
        _conn.execute(sqlQuery)
        _conn.commit()

DEPLOYMENT_TRACKER_TBL_NAME = "VERSION_TRACKER"

class __DeploymentTrackerStorage(Storage):
    def __init__(self):
        self._initTable(DEPLOYMENT_TRACKER_TBL_NAME,"VERSION TEXT NOT NULL")

def _trackDeployment():
    deploymenTrackerStorage = __DeploymentTrackerStorage()
    row = deploymenTrackerStorage.fetchOne("SELECT * FROM {0}".format(DEPLOYMENT_TRACKER_TBL_NAME));
    if row is None:
        _trackDeploymentIfVersionChange(deploymenTrackerStorage, None)
    else:
        _trackDeploymentIfVersionChange(deploymenTrackerStorage, row["VERSION"])

def _trackDeploymentIfVersionChange(deploymenTrackerStorage, existingVersion):
    # Get version and repository URL from 'setup.py'
    version = None
    repo_url = None
    try:
        app = get_distribution("pixiedust")
        version = app.version
        repo_url = PIXIEDUST_REPO_URL
        notebook_tenant_id = os.environ.get("NOTEBOOK_TENANT_ID")
        notebook_kernel = os.environ.get("NOTEBOOK_KERNEL")
        # save last tracked version in the db
        if existingVersion is None:
            deploymenTrackerStorage.insert("INSERT INTO {0} (VERSION) VALUES ('{1}')".format(DEPLOYMENT_TRACKER_TBL_NAME,version))
        else:
            deploymenTrackerStorage.update("UPDATE {0} SET VERSION='{1}'".format(DEPLOYMENT_TRACKER_TBL_NAME,version))
        # if version has changed then track with deployment tracker
        if existingVersion is None or existingVersion != version:
            myLogger.info("Change in version detected: {0} -> {1}.".format(existingVersion,version))
            if existingVersion is None:
                printWithLogo("Pixiedust version {0}".format(version))
            else:
                printWithLogo("Pixiedust version upgraded from {0} to {1}".format(existingVersion,version))
            # create dictionary and register
            event = dict()
            event['date_sent'] = time.strftime("%Y-%m-%dT%H:%M:%S.000Z", time.gmtime())
            if version is not None:
                event['code_version'] = version
            if repo_url is not None:
                event['repository_url'] = repo_url
            if notebook_tenant_id is not None:
                event['notebook_tenant_id'] = notebook_tenant_id
            if notebook_kernel is not None:
                event['notebook_kernel'] = notebook_kernel
            event['runtime'] = 'python'
            # Create and format request to Deployment Tracker
            url = 'https://deployment-tracker.mybluemix.net/api/v1/track'
            headers = {'content-type': "application/json"}
            response = post(url, data=json.dumps(event), headers=headers)
        else:
            myLogger.info("No change in version: {0} -> {1}.".format(existingVersion,version))
            printWithLogo("Pixiedust version {0}".format(version)) 
    except:
        myLogger.error("Error registering with deployment tracker:\n" + str(sys.exc_info()[0]) + "\n" + str(sys.exc_info()[1]))


# -------------------------------------------------------------------------------
# Copyright IBM Corp. 2017
# 
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
# http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -------------------------------------------------------------------------------

from pixiedust.display.chart.renderers import PixiedustRenderer
from .googleBaseDisplay import GoogleBaseDisplay

import numpy as np
import pixiedust

myLogger = pixiedust.getLogger(__name__)

@PixiedustRenderer(id="mapView")
class MapViewDisplay(GoogleBaseDisplay):

    def supportsKeyFieldLabels(self, handlerId):
        return False
    
    def getPreferredDefaultValueFieldCount(self, handlerId):
        return 1

    def getDefaultKeyFields(self, handlerId, aggregation):
        fields = self._getDefaultKeyFields()
        if (len(fields) > 0):
            return fields
        else:
            return super(MapViewDisplay, self).getDefaultKeyFields(handlerId, aggregation) # no relevant fields found - defer to superclass
    
    def getChartContext(self, handlerId):
        diagTemplate = GoogleBaseDisplay.__module__ + ":mapViewOptionsDialogBody.html"
        return (diagTemplate, {})
    
    def canRenderChart(self):
        keyFields = self.getKeyFields()
        if ((keyFields is not None and len(keyFields) > 0) or len(self._getDefaultKeyFields()) > 0):
            return (True, None)
        else:
            return (False, "No location field found ('country', 'province', 'state', 'city', or 'latitude'/'longitude').<br>Use the Chart Options dialog to specify a location field.")

    def doRenderChart(self):
        keyFields = self.getKeyFields()
        # keyFieldLabels = self.getKeyFieldLabels()
        valueFields = self.getValueFields()
        latLong = self.dataHandler.isNumericField(keyFields[0])
        apikey = self.options.get("googlemapapikey")

        if self.options.get("mapRegion") is None:
            if keyFields[0].lower() == "state":
                self.options["mapRegion"] = "US"
            else:    
                self.options["mapRegion"] = "world"

        if self.options.get("mapDisplayMode") is None:
            if latLong:
                self.options["mapDisplayMode"] = "markers"
            else:
                self.options["mapDisplayMode"] = "region"

        if self.options.get("mapDisplayMode") != "region" and (apikey is None or len(apikey)<5):
            return self.renderTemplate("noapikey.html")
            
        if self.options["mapRegion"] == "US":
            self.options["mapResolution"] = "provinces"
        else:
            self.options["mapResolution"] = "countries"

        df = self.getWorkingPandasDataFrame()
        colData = str(df.columns.values.tolist())
        valData = str(df.values.tolist()).encode('utf-8')
        mapData = "[" + valData.replace('[', (colData + ", "), 1)

        self.options["mapData"] = mapData.replace("'",'"').replace('[u"', '["').replace(', u"', ', "')
        self._addScriptElement("https://www.gstatic.com/charts/loader.js")
        if apikey is not None and len(apikey)>5:
            self._addScriptElement("https://maps.googleapis.com/maps/api/js?key={0}".format(apikey))
        self._addScriptElement("https://www.google.com/jsapi", callback=self.renderTemplate("mapView.js"))
        return self.renderTemplate("mapView.html")

    def _getDefaultKeyFields(self):
        for field in self.entity.schema.fields:
            if field.name.lower() == 'country' or field.name.lower() == 'province' or field.name.lower() == 'state' or field.name.lower() == 'city':
                return [field.name]
        # check for lat/long
        latLongFields = []
        for field in self.entity.schema.fields:
            if field.name.lower() == 'lat' or field.name.lower() == 'latitude':
                latLongFields.append(field.name)
        for field in self.entity.schema.fields:
            if field.name.lower() == 'lon' or field.name.lower() == 'long' or field.name.lower() == 'longitude':
                latLongFields.append(field.name)
        if (len(latLongFields) == 2):
            return latLongFields
        return []



# -------------------------------------------------------------------------------
# Copyright IBM Corp. 2016
# 
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
# http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -------------------------------------------------------------------------------

import warnings
from IPython.core.getipython import get_ipython

__all__ = ['addDisplayRunListener', 'display']

#Make sure that matplotlib is running inline
with warnings.catch_warnings():
    warnings.simplefilter("ignore")
    try:
        get_ipython().run_line_magic("matplotlib", "inline")
    except NameError:
        #IPython not available we must be in a spark executor
        pass

displayRunListeners = []

def addDisplayRunListener(listener):
    global displayRunListeners
    displayRunListeners.append(listener)

from .display import *
from .chart import *
from .graph import *
from .table import *
from .download import *
from .datahandler import getDataHandler
from pixiedust.utils.printEx import *
import traceback
import uuid
import pixiedust
from six import string_types

myLogger=pixiedust.getLogger(__name__ )

def display(entity, **kwargs):
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")

        #todo: use ConverterRegistry
        def toPython(entity):
            from py4j.java_gateway import JavaObject
            if entity is None or not isinstance(entity, JavaObject):
                return entity

            clazz = entity.getClass().getName()
            if clazz == "org.apache.spark.sql.Dataset":
                entity = entity.toDF()
                clazz = "org.apache.spark.sql.DataFrame"

            if clazz == "org.apache.spark.sql.DataFrame":
                from pyspark.sql import DataFrame, SQLContext
                from pyspark import SparkContext
                entity = DataFrame(entity, SQLContext(SparkContext.getOrCreate(), entity.sqlContext()))

            return entity

        callerText = traceback.extract_stack(limit=2)[0][3]

        pr = None
        try:
            if "cell_id" in kwargs and "showchrome" not in kwargs and "handlerId" in kwargs:
                if "gen_tests" in kwargs:
                    #remove gen_tests from command line
                    import re
                    m = re.search(",\\s*gen_tests\\s*=\\s*'((\\\\'|[^'])*)'", str(callerText), re.IGNORECASE)
                    if m is not None:
                        callerText = callerText.replace(m.group(0),"")
                    #generate new prefix
                    p = re.search(",\\s*prefix\\s*=\\s*'((\\\\'|[^'])*)'", str(callerText), re.IGNORECASE)
                    if p is not None:
                        prefix = ''.join([",prefix='", str(uuid.uuid4())[:8], "'"])
                        callerText = callerText.replace(p.group(0), prefix)
                    get_ipython().set_next_input(callerText)
                if "profile" in kwargs:
                    import cProfile
                    pr = cProfile.Profile()
                    pr.enable()

            scalaKernel = False
            if callerText is None or callerText == "" and hasattr(display, "fetchEntity"):
                callerText, entity = display.fetchEntity(entity)
                entity = toPython(entity)
                scalaKernel = True

            #get a datahandler and displayhandler for this entity
            dataHandler = getDataHandler(kwargs, entity)
            selectedHandler = getSelectedHandler(kwargs, entity, dataHandler)

            #notify listeners of a new display Run
            for displayRunListener in displayRunListeners:
                displayRunListener(entity, kwargs)
            
            #check if we have a job monitor id
            from pixiedust.utils.sparkJobProgressMonitor import progressMonitor
            if progressMonitor:
                progressMonitor.onDisplayRun(kwargs.get("cell_id"))
            
            myLogger.debug("Creating a new display handler with options {0}: {1}".format(kwargs, selectedHandler))
            displayHandler = selectedHandler.newDisplayHandler(kwargs,entity)
            if displayHandler is None:
                printEx("Unable to obtain handler")
                return
            
            displayHandler.handlerMetadata = selectedHandler
            displayHandler.dataHandler = dataHandler
            displayHandler.callerText = callerText
            if scalaKernel:
                displayHandler.scalaKernel = True

            if displayHandler.callerText is None:
                printEx("Unable to get entity information")
                return

            displayHandler.render()
        finally:
            if pr is not None:
                import pstats, StringIO
                pr.disable()
                s = StringIO.StringIO()
                ps = pstats.Stats(pr, stream=s).sort_stats('cumulative')
                ps.print_stats()
                myLogger.debug(s.getvalue())


# -------------------------------------------------------------------------------
# Copyright IBM Corp. 2017
# 
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
# http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -------------------------------------------------------------------------------

from pixiedust.display.chart.renderers import PixiedustRenderer
from pixiedust.display.chart.colorManager import Colors
from .bokehBaseDisplay import BokehBaseDisplay
from pixiedust.utils import Logger
from bokeh.charts import Line
from bokeh.plotting import figure
import numpy as np
import sys

@PixiedustRenderer(id="lineChart")
@Logger()
class LineChartRenderer(BokehBaseDisplay):
    def isSubplot(self):
        return self.options.get("lineChartType", "grouped") == "subplots"

    def getExtraFields(self):
        if not self.isSubplot() and len(self.getValueFields())>1:
            #no clusterby if we are grouped and multiValueFields
            return []
    
        clusterby = self.options.get("clusterby")
        return [clusterby] if clusterby is not None else []

    def createBokehChart(self):
        keyFields = self.getKeyFields()
        valueFields = self.getValueFields()
        clusterby = self.options.get("clusterby")
        subplots = self.isSubplot()
        workingPDF = self.getWorkingPandasDataFrame().copy()

        for index, row in workingPDF.iterrows():
            for k in keyFields:
                if isinstance(row[k], str if sys.version >= '3' else basestring):
                    row[k] = row[k].replace(':', '.')
            workingPDF.loc[index] = row

        charts=[]
        if clusterby is not None and (subplots or len(valueFields)<=1):
            subplots = subplots if len(valueFields)==1 or subplots else False
            for j, valueField in enumerate(valueFields):
                pivot = workingPDF.pivot(
                    index=keyFields[0], columns=clusterby, values=valueField
                )

                if not subplots:
                    fig = figure(x_range=pivot.index.values.tolist() if not np.issubdtype(pivot.index.dtype, np.number) else None)
                    charts.append(fig)
                for i,col in enumerate(pivot.columns[:10]): #max 10
                    if subplots:
                        charts.append( 
                            Line(
                                pivot[col].values, color=Colors.hexRGB( 1.*i/2 ), ylabel=valueField, xlabel=keyFields[0], legend=False, 
                                title="{0} = {1}".format(clusterby, pivot.columns[i])
                            )
                        )
                    else:
                        xValues = pivot.index.values.tolist()
                        if not np.issubdtype(pivot.index.dtype, np.number):
                            xValues = range(1, len(xValues)+1)
                        fig.line(x = xValues, y = pivot[col].values, color = Colors.hexRGB( 1.*i/2 ), legend=str(col) if self.showLegend() else None)
        else:
            if subplots:
                for i,valueField in enumerate(valueFields):
                    charts.append(Line(workingPDF, x = keyFields[0], y=valueField, color = Colors.hexRGB( 1.*i/2 ), legend=self.showLegend(), plot_width=int(800/len(valueFields))))
            else:
                charts.append(Line(workingPDF, x = keyFields[0], y=valueFields, color=valueFields, legend=self.showLegend()))

            if clusterby is not None:
                self.addMessage("Warning: 'Cluster By' ignored when grouped option with multiple Value Fields is selected")
        return charts


# -------------------------------------------------------------------------------
# Copyright IBM Corp. 2017
# 
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
# http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -------------------------------------------------------------------------------
from pixiedust.utils.dataFrameAdapter import PandasDataFrameAdapter
import pixiedust.utils.dataFrameMisc as dataFrameMisc
import numpy as np
from pixiedust.utils import Logger

@Logger()
class PandasDataFrameDataHandler(object):
    def __init__(self, options, entity):
        self.options = options
        self.entity = entity

    def getFieldNames(self, expandNested=False):
        return dataFrameMisc.getFieldNames(PandasDataFrameAdapter(self.entity), expandNested)

    def isNumericField(self, fieldName):
        for y in self.entity.columns:
            if y == fieldName:
                return self.entity[y].dtype == np.float64 or self.entity[y].dtype == np.int64
        raise ValueError("Column {} does not existing in the dataframe".format(fieldName))

    def isNumericType(self, field):
        return dataFrameMisc.isNumericType(field.dataType)

    def isStringField(self, fieldName):
        return dataFrameMisc.isStringField(PandasDataFrameAdapter(self.entity), fieldName)

    def isStringType(self, field):
        return dataFrameMisc.isStringType(field.dataType)

    @property
    def schema(self):
        return PandasDataFrameAdapter(self.entity).schema

    """
        Return a cleaned up Pandas Dataframe that will be used as working input to the chart
    """
    def getWorkingPandasDataFrame(self, xFields, yFields, extraFields=[], aggregation=None, maxRows = 100):
        if xFields is None or len(xFields)==0:
            #swap the yFields with xFields
            xFields = yFields
            yFields = []
            aggregation = None

        extraFields = [a for a in extraFields if a not in xFields and a not in yFields]
        workingDF = self.entity[xFields + extraFields + yFields]

        if aggregation and len(yFields)>0:
            aggMapper = {"SUM":"sum", "AVG": "mean", "MIN": "min", "MAX": "max"}
            aggFn = aggMapper.get(aggregation, "count")
            workingDF = workingDF.groupby(extraFields + xFields).agg(aggFn).reset_index()

        workingDF = workingDF.dropna()
        count = len(workingDF.index)
        if count > maxRows:
            workingDF = workingDF.sample(frac=(float(maxRows) / float(count)),replace=False)
        
        #sort by xFields
        workingDF.sort_values(extraFields + xFields, inplace=True)
        
        return workingDF


# -------------------------------------------------------------------------------
# Copyright IBM Corp. 2016
# 
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
# http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -------------------------------------------------------------------------------

from pixiedust.display.chart.renderers import PixiedustRenderer
from .matplotlibBaseDisplay import MatplotlibBaseDisplay
from pixiedust.display.chart.colorManager import Colors
import matplotlib.pyplot as plt
import mpld3
import numpy as np
import pixiedust

myLogger = pixiedust.getLogger(__name__)

@PixiedustRenderer(id="histogram")
class HistogramDisplay(MatplotlibBaseDisplay):

    def supportsAggregation(self, handlerId):
        return False

    # TODO: add support for keys
    def supportsKeyFields(self, handlerId):
        return False
    
    def getPreferredDefaultValueFieldCount(self, handlerId):
        return 1

    # no keys by default
    def getDefaultKeyFields(self, handlerId, aggregation):
        return []

    def getNumFigures(self):
        return len(self.getValueFields()) if self.isSubplots() else 1

    def isSubplots(self):
        return len(self.getValueFields()) > 1 and self.options.get("histoChartType", "stacked") == "subplots"

    def matplotlibRender(self, fig, ax):
        stacked = len(self.getValueFields()) > 1 and self.options.get("histoChartType", "stacked") == "stacked"
        subplots = self.isSubplots()
        binsize = int(self.options.get('binsize', 10))

        def plot(ax, valueField=None, color=None):
            data = self.getWorkingPandasDataFrame() if valueField is None else self.getWorkingPandasDataFrame()[valueField]
            data.plot(
                kind="hist", stacked=stacked, ax=ax, bins=binsize, legend=self.showLegend(), x = valueField,
                label=valueField,color = color, colormap = None if color else Colors.colormap
            )

        if subplots:
            for j, valueField in enumerate(self.getValueFields()):
                plot(self.getAxItem(ax, j), valueField, Colors.colormap(1.*j/2))
        else:
            plot(ax)


# -------------------------------------------------------------------------------
# Copyright IBM Corp. 2017
# 
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
# http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -------------------------------------------------------------------------------
import math

def append(displayObject, arr, option):
    if displayObject.acceptOption(option["name"]):
        arr.append(option)

def clusterBy(displayObject):
    return { 
        'name': 'clusterby',
        'description': 'Cluster By',
        'refresh': True,
        'metadata': {
            'type': "dropdown",
            'values': ["None"] + sorted([f for f in displayObject.getFieldNames() if f not in displayObject.getKeyFields() and f not in displayObject.getValueFields()]),
            'default': ""
        },
        'validate': lambda option:\
            (option in displayObject.getFieldNames() and option not in displayObject.getKeyFields() and option not in displayObject.getValueFields(),\
             "Cluster By value is already used in keys or values for this chart")
    }

def barChart(displayObject):
    options = []
    options.append(clusterBy(displayObject))

    if not hasattr(displayObject, 'no_orientation') or displayObject.no_orientation is not True:
        options.append({
            'name': 'orientation',
            'description': 'Orientation',
            'metadata': {
                'type': 'dropdown',
                'values': ['vertical', 'horizontal'],
                'default': "vertical"
            }
        })

    if displayObject.options.get("clusterby") != None or len(displayObject.getValueFields()) > 1:
        options.append({
            'name': 'charttype',
            'description': 'Type',
            'metadata': {
                'type': 'dropdown',
                'values': ['grouped', 'stacked', 'subplots'],
                'default': "grouped"
            }
        })

    options.append({
        'name': 'legend',
        'description': 'Show legend',
        'metadata': {
            'type': 'checkbox',
            'default': "true"
        }
    })

    return options

def lineChart(displayObject):
    options = []

    options.append(clusterBy(displayObject))

    if displayObject.options.get("clusterby") != None or len(displayObject.getValueFields()) > 1:
        options.append({
            'name': 'lineChartType',
            'description': 'Type',
            'metadata': {
                'type': 'dropdown',
                'values': ['grouped', 'subplots'],
                'default': "grouped"
            }
        })

    options.append({
        'name': 'legend',
        'description': 'Show legend',
        'metadata': {
            'type': 'checkbox',
            'default': "false"
        }
    })

    options.append({
        'name': 'logx',
        'description': 'log scale on x',
        'metadata': {
            'type': 'checkbox',
            'default': "false"
        }
    })
    options.append({
        'name': 'logy',
        'description': 'log scale on y',
        'metadata': {
            'type': 'checkbox',
            'default': "false"
        }
    })
    return options

def histogram(displayObject):
    options = []
    if len(displayObject.getValueFields()) > 1:
        append(displayObject, options, {
            'name': 'histoChartType',
            'description': 'Type',
            'metadata': {
                'type': 'dropdown',
                'values': ['stacked', 'subplots'],
                'default': "stacked"
            }
        })
    count = len(displayObject.getWorkingPandasDataFrame().index)
    options.append({
        'name': 'binsize',
        'description': 'Bin size',
        'metadata': {
            'type': 'slider',
            'max': int(max(math.ceil(count / 2), 4)),
            'min': int(max(math.floor(count / 20), 2)),
            'default': int(max(math.ceil(count / 4), 3))
        }
    })
    return options

commonOptions = {}
for f in [barChart,lineChart,histogram]:
    commonOptions.update({f.__name__:f})



# -------------------------------------------------------------------------------
# Copyright IBM Corp. 2017
# 
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
# http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -------------------------------------------------------------------------------

from pixiedust.display.chart.renderers import PixiedustRenderer
from pixiedust.display.chart.renderers.baseChartDisplay import commonChartOptions
from pixiedust.display.chart.colorManager import Colors
from .bokehBaseDisplay import BokehBaseDisplay
from pixiedust.utils import Logger
from bokeh.charts import Histogram, show
from bokeh.layouts import row
import math

@PixiedustRenderer(id="histogram")
@Logger()
class HistogramRenderer(BokehBaseDisplay):
    
    def supportsAggregation(self, handlerId):
        return True

    def supportsKeyFields(self, handlerId):
        return False
    
    def getPreferredDefaultValueFieldCount(self, handlerId):
        return 1

    def getDefaultKeyFields(self, handlerId, aggregation):
        return []

    def acceptOption(self, optionName):
        if optionName == 'histoChartType':
            return False
        return True

    @commonChartOptions
    def getChartOptions(self):
        return [
            { 'name': 'color',
              'metadata': {
                    'type': "dropdown",
                    'values': ["None"] + self.getFieldNames(),
                    'default': ""
                }
            }
        ]

    def canRenderChart(self):
        return (True, None)

    def getExtraFields(self):
        color = self.options.get("color")
        return [color] if color is not None else []
    
    def createBokehChart(self):
        plotwidth = int(self.getPreferredOutputWidth()/2)
        plotheight = int(self.getPreferredOutputHeight() * 0.75)
        binsize = int(self.options.get('binsize', 10))  
        valueFields = self.getValueFields()
        histograms = []
        if len(valueFields) != 1:
            for i,valueField in enumerate(valueFields):
                color = self.options.get("color")
                if color is None:
                    color = Colors.hexRGB( 1.*i/2 )
                histograms.append(Histogram(
                    self.getWorkingPandasDataFrame(), values=valueField, plot_width=plotwidth, plot_height=plotheight,bins=binsize, 
                    color=color, xgrid=True, ygrid=True, ylabel='Frequency')
                )
            return histograms
        else:
            return Histogram(self.getWorkingPandasDataFrame(), values=self.getValueFields()[0],
            bins=binsize, color=self.options.get("color"), 
            xgrid=True, ygrid=True, ylabel='Frequency')


# -------------------------------------------------------------------------------
# Copyright IBM Corp. 2016
# 
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
# http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -------------------------------------------------------------------------------

from __future__ import print_function
import sqlite3
import sys
import platform
import urllib
import os
from pixiedust.utils.storage import *
from pixiedust.utils.printEx import *
from pyspark import SparkContext
from .package import Package
from .downloader import Downloader, RequestException
import six

PACKAGES_TBL_NAME="SPARK_PACKAGES"

class __PackageStorage(Storage):
    def __init__(self):
        self._initTable( PACKAGES_TBL_NAME,
        '''
            GROUPID        TEXT  NOT NULL,
            ARTIFACTID     TEXT  NOT NULL,
            VERSION        TEXT  NOT NULL,
            FILEPATH       TEXT  NOT NULL,
            BASE           TEXT,
            PRIMARY KEY (GROUPID, ARTIFACTID)
        ''')

packageStorage = __PackageStorage()

class PackageManager(object):
    DOWNLOAD_DIR=os.environ.get("PIXIEDUST_HOME", os.path.expanduser('~')) + "/data/libs"
    
    def __init__(self):
        if not os.path.exists(self.DOWNLOAD_DIR):
            os.makedirs(self.DOWNLOAD_DIR)

    def _toPackage(self, package):
        if isinstance(package, six.string_types):
            package=Package.fromPackageIdentifier(package)
        return package
    
    def installPackage(self, package, base=None, sc=None):
        package = self._toPackage(package)
        #Test if we already have a version installed
        res=self.fetchPackage(package)
        fileLoc=None
        if res:
            fileLoc=res[1]
            print("Package already installed: {0}".format(str(package)))
        else:
            #download package
            packs=[package]
            def _doDownload(d):
                package=packs[0]
                if not package.version or package.version=='0':
                    package.version = d.resolver._find_latest_version_available(package)
                fileLoc = package.getFilePath(self.DOWNLOAD_DIR)
                if os.path.isfile(fileLoc):
                    os.remove(fileLoc)
                results = d.download(package,filename=self.DOWNLOAD_DIR)
                if not results[1]:
                    raise Exception("Error downloading package {0}".format(str(package)))
                else:
                    package=results[0]
                    print("Package {0} downloaded successfully".format(str(package)))
                    printEx("Please restart Kernel to complete installation of the new package",PrintColors.RED)
                fileLoc=self.storePackage(package,base)
                return fileLoc
            
            try:
                fileLoc=_doDownload(Downloader(base) if base is not None else Downloader())
            except RequestException as e:
                #try another base
                try:
                    fileLoc=_doDownload(Downloader("http://dl.bintray.com/spark-packages/maven"))
                except RequestException as e:
                    print("Unable to install package {0}".format(e.msg))
                    raise
            except:
                print(str(sys.exc_info()[1]))
                raise
        if sc is None:
            sc = SparkContext.getOrCreate()
            
        if sc:
            #convert to file uri for windows platform
            if platform.system()=='Windows':
                fileLoc="file://" + urllib.pathname2url(fileLoc)
            sc.addPyFile(fileLoc)
            
        return package
        
    def uninstallPackage(self, package):
        results = self.fetchPackage(package)
        if results is None:
            print("Package is not installed: {0}".format(str(package)))
        else:
            self._deletePackage(results[0],results[1])
    
    def hasPackage(self, package):
        return self.fetchPackage(package) is not None
        
    def fetchPackage(self, package ):
        package = self._toPackage(package)
        return packageStorage.fetchOne("""
                SELECT * from {0} WHERE GROUPID='{1}' and ARTIFACTID='{2}'
            """.format(PACKAGES_TBL_NAME,package.group_id,package.artifact_id), 
            lambda row: (Package(row["GROUPID"],row["ARTIFACTID"],row["VERSION"]),row["FILEPATH"])
        )
                
    def printAllPackages(self):
        self.visitAll(lambda row: print("{0}:{1}:{2} => {3}".format(row["GROUPID"],row["ARTIFACTID"],row["VERSION"],row["FILEPATH"])))
        
    def visitAll(self, walker):
        packageStorage.execute("""
                SELECT * FROM {0}
            """.format(PACKAGES_TBL_NAME),
            walker
        )
        
    def storePackage(self, package,base=None):
        package = self._toPackage(package)
        fileLoc=package.getFilePath(self.DOWNLOAD_DIR)
        packageStorage.insert("""
            INSERT INTO {0} (GROUPID,ARTIFACTID,VERSION,BASE,FILEPATH)
            VALUES ('{1}','{2}','{3}','{4}','{5}')
        """.format(
                PACKAGES_TBL_NAME,
                package.group_id,
                package.artifact_id,
                package.version,
                base if base is not None else "",
                fileLoc
            )
        )
        print("Successfully added package {0}".format(str(package)))
        return fileLoc
        
    def _deletePackage(self,package,filePath):
        package = self._toPackage(package)
        rowDeleted = packageStorage.delete("""
            DELETE FROM {0} WHERE GROUPID='{1}' AND ARTIFACTID='{2}'
        """.format(
                PACKAGES_TBL_NAME,
                package.group_id,
                package.artifact_id
            )
        )
        if rowDeleted==0:
            print("Package {0} not deleted because it doesn't exist".format(str(package)))
        else:
            print("Successfully deleted package {0}".format(str(package)))
            
        if filePath is not None:
            os.remove(filePath)
    


# -------------------------------------------------------------------------------
# Copyright IBM Corp. 2016
# 
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
# http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -------------------------------------------------------------------------------

import os
from . import storage
import pkg_resources
import binascii
import shutil
from . import pdLogging

storage._initStorage()

#Misc helper methods
def fqName(entity):
    return (entity.__module__ + "." if hasattr(entity, "__module__") else "") + entity.__class__.__name__

#init scala bridge, make sure that correct pixiedust.jar is installed

jarDirPath = os.environ.get("PIXIEDUST_HOME", os.path.expanduser('~')) + "/data/libs/"
jarFilePath = jarDirPath + "pixiedust.jar"

dir = os.path.dirname(jarDirPath)
if not os.path.exists(dir):
    os.makedirs(dir)

def installPixiedustJar():
    with pkg_resources.resource_stream(__name__, "resources/pixiedust.jar") as resJar:
        with open( jarFilePath, 'wb+' ) as installedJar:
            shutil.copyfileobj(resJar, installedJar)
            print("Pixiedust runtime updated. Please restart kernel")

copyFile = True
if os.path.isfile(jarFilePath):
    with open( jarFilePath, 'rb' ) as installedJar:
        installedCRC = binascii.crc32( installedJar.read() )
        with pkg_resources.resource_stream(__name__, "resources/pixiedust.jar") as resJar:
            copyFile = installedCRC != binascii.crc32( resJar.read() )

if copyFile:
    installPixiedustJar()

def checkVersion():
    import json
    try:
        from urllib.request import urlopen
    except ImportError:
        from urllib2 import urlopen

    def printHTML(html):
        from IPython.display import display, HTML
        display(HTML(html))
    
    try:
        response = urlopen("https://pypi.python.org/pypi/pixiedust/json", timeout=3)
        latest = json.loads(response.read())["info"]["version"]
        version = pkg_resources.get_distribution("pixiedust").parsed_version._version.release
        version = ".".join([str(p) for p in version])
        if latest != version:
            printHTML("<div>Warning: You are not running the latest version of PixieDust. Current is {0}, Latest is {1}</div>".format(version, latest))
            printHTML("""
                <div>Please copy and run the following command in a new cell to upgrade: <span style="background-color:#ececec;font-family:monospace;padding:0 5px">{0}</span></div>
            """.format("!pip install --user --upgrade pixiedust"))
    except Exception as e:
        print("Unable to check latest version {0}".format(e))

"""
Helper decorator that automatically cache results of a class method into a field
"""
def cache(fieldName):
    def outer(func):
        if fieldName == func.__name__:
            raise AttributeError("cached fieldName cannot have the same name as the function: {}".format(fieldName))
        def inner(cls, *args, **kwargs):
            if hasattr(cls, fieldName) and getattr(cls, fieldName) is not None:
                return getattr(cls, fieldName)
            retValue = func(cls, *args, **kwargs)
            setattr(cls, fieldName, retValue)
            return retValue
        return inner
    return outer

"""
Helper decorator that automatically add Logging capability to a class
"""
class Logger(object):
    def __call__(self, cls, *args, **kwargs):
        if not hasattr(cls, "myLogger"):
            cls.myLogger = pdLogging.getLogger(cls.__module__ + "." + cls.__name__)
            cls.debug = cls.myLogger.debug
            cls.warn = cls.myLogger.warn
            cls.error = cls.myLogger.error
            cls.info = cls.myLogger.info
            cls.exception = cls.myLogger.exception
        return cls



# -------------------------------------------------------------------------------
# Copyright IBM Corp. 2017
# 
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
# http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -------------------------------------------------------------------------------

from pixiedust.display.chart.renderers import PixiedustRenderer
from .mapBoxBaseDisplay import MapBoxBaseDisplay
from pixiedust.utils import cache
from pixiedust.utils import Logger
import json 

@PixiedustRenderer(id="mapView")
@Logger()
class MapViewDisplay(MapBoxBaseDisplay):
    def isMap(self, handlerId):
        return True

    def supportsAggregation(self, handlerId):
        return True

    def getDefaultAggregation(self, handlerId):
        return "SUM"

    def supportsLegend(self, handlerId):
        return True

    def getPreferredDefaultKeyFieldCount(self, handlerId):
        return 2

    def getPreferredDefaultValueFieldCount(self, handlerId):
        return 1

    def canRenderChart(self):
        keyFields = self.getKeyFields()
        if ((keyFields is not None and len(keyFields) > 0) or len(self._getDefaultKeyFields()) > 0):
            return (True, None)
        else:
            return (False, "No location field found ('latitude'/'longitude', 'lat/lon', 'y/x').<br>Use the Chart Options dialog to specify location fields.")

    def getChartContext(self, handlerId):
        diagTemplate = MapBoxBaseDisplay.__module__ + ":mapViewOptionsDialogBody.html"
        return (diagTemplate, {})
    
    def doRenderChart(self):
        mbtoken = self.options.get("mapboxtoken")
        if not mbtoken or len(mbtoken)<5:
            return self.renderTemplate("noaccesstoken.html")

        df = self.getWorkingPandasDataFrame()

        keyFields = self.getKeyFields()
        lonFieldIdx = 0
        latFieldIdx = 1
        if keyFields[0] == self.getLatField(): 
            lonFieldIdx = 1
            latFieldIdx = 0
        valueFields = self.getValueFields()
        valueFieldIdxs = []
        for j, valueField in enumerate(valueFields):
            valueFieldIdxs.append(df.columns.get_loc(valueField))
        

        min = [df[keyFields[lonFieldIdx]].min(), df[keyFields[latFieldIdx]].min()]
        max = [df[keyFields[lonFieldIdx]].max(), df[keyFields[latFieldIdx]].max()]

        # Transform the data into GeoJSON for use in the Mapbox client API
        pygeojson = {'type':'FeatureCollection', 'features':[]}

        for row in df.itertuples():
            feature = {'type':'Feature',
                        'properties':{},
                        'geometry':{'type':'Point',
                                    'coordinates':[]}}
            feature['geometry']['coordinates'] = [row[lonFieldIdx+1], row[latFieldIdx+1]]
            for idx, valueFieldIdx in enumerate(valueFieldIdxs):
                feature['properties'][valueFields[idx]] = row[valueFieldIdx+1]
            pygeojson['features'].append(feature)

        self.options["mapBounds"] = json.dumps([min,max])
        self.options["mapData"] = json.dumps(pygeojson)

        paint = {'circle-radius':12,'circle-color':'#ff0000'}
        bins = []

        if len(valueFields) > 0:
            mapValueField = valueFields[0]
            self.options["mapValueField"] = mapValueField

        if not self.options.get("kind"): 
            self.options["kind"] = "choropleth"
        # if there's a numeric value field paint the data as a choropleth map
        if self.options.get("kind") != "simple" and len(valueFields) > 0:
            minval = df[valueFields[0]].min()
            maxval = df[valueFields[0]].max()
            bins = [ (minval,'#ffffcc'), (df[valueFields[0]].quantile(0.25),'#a1dab4'), (df[valueFields[0]].quantile(0.5),'#41b6c4'), (df[valueFields[0]].quantile(0.75),'#2c7fb8'), (maxval,'#253494') ]
            paint['circle-opacity'] = 0.85
            paint['circle-color'] = {"property":mapValueField}
            paint['circle-color']['stops'] = []
            for bin in bins: 
                paint['circle-color']['stops'].append( [bin[0], bin[1]] )
        self.options["mapStyle"] = json.dumps(paint)
        w = self.getPreferredOutputWidth()
        h = self.getPreferredOutputHeight()
        body = self.renderTemplate("mapView.html", bins=bins, prefwidth=w, prefheight=h)
        return self.renderTemplate("iframesrcdoc.html", body=body, prefwidth=w, prefheight=h)

    def isLatLonChart(self):
        llnames = ['lat','latitude','y','lon','long','longitude','x']
        isll = True;
        keyFields = self.getKeyFields()
        if ((keyFields is not None) and len(keyFields) == 2):
            for field in keyFields:
                if field.lower() not in llnames:
                    isll = False
        return isll;

    def getLatField(self):
        names = ['lat','latitude','y']
        keyFields = self.getKeyFields()
        if (keyFields is not None):
            for field in keyFields:
                if field.lower() in names:
                    return field
        return None;

    def getLonField(self):
        names = ['lon','long','longitude','x']
        keyFields = self.getKeyFields()
        if (keyFields is not None):
            for field in keyFields:
                if field.lower() in names:
                    return field
        return None;

    def _getDefaultKeyFields(self):
        # check for lat/long
        latLongFields = []
        for field in self.entity.schema.fields:
            if field.name.lower() == 'lat' or field.name.lower() == 'latitude' or field.name.lower() == 'y':
                latLongFields.append(field.name)
        for field in self.entity.schema.fields:
            if field.name.lower() == 'lon' or field.name.lower() == 'long' or field.name.lower() == 'longitude' or field.name.lower() == 'x':
                latLongFields.append(field.name)
        if (len(latLongFields) == 2):
            return latLongFields
        # if we get here, look for an address field
        for field in self.entity.schema.fields:
            if field.name.lower() == 'address':
                return latLongFields.append(field.name)
        return []



# -------------------------------------------------------------------------------
# Copyright IBM Corp. 2017
# 
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
# http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -------------------------------------------------------------------------------

from pixiedust.display.chart.renderers import PixiedustRenderer
from .matplotlibBaseDisplay import MatplotlibBaseDisplay
from pixiedust.display.chart.colorManager import Colors

@PixiedustRenderer(id="scatterPlot")
class ScatterPlotDisplay(MatplotlibBaseDisplay):
    
	def supportsAggregation(self, handlerId):
		return False
    
	def canRenderChart(self):
		valueFields = self.getValueFields()
		if len(valueFields) != 1:
			return (False, "Can only specify one Value Field")

		#Verify that all key field are numericals
		for keyField in self.getKeyFields():
			if not self.dataHandler.isNumericField(keyField):
				return (False, "Column {0} is not numerical".format(keyField))
		
		return (True, None)

	def getPreferredDefaultValueFieldCount(self, handlerId):
		return 2

	def matplotlibRender(self, fig, ax):
		keyFields = self.getKeyFields()
		for i,keyField in enumerate(keyFields):
			self.getWorkingPandasDataFrame().plot(kind='scatter', x=keyField, y=self.getValueFields()[0], 
				label=keyField, ax=ax, color=Colors[1.*i/len(keyFields)], figsize=(8, 8))


# -------------------------------------------------------------------------------
# Copyright IBM Corp. 2017
# 
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
# http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -------------------------------------------------------------------------------

from pixiedust.display.chart.renderers import PixiedustRenderer
from pixiedust.display.chart.colorManager import Colors
from .matplotlibBaseDisplay import MatplotlibBaseDisplay
import matplotlib.pyplot as plt
import numpy as np
from pixiedust.utils import Logger

@PixiedustRenderer(id="lineChart")
@Logger()
class LineChartDisplay(MatplotlibBaseDisplay):

    def getNumFigures(self):
        return len(self.getValueFields()) if self.isSubplot() else 1

    def isSubplot(self):
        return self.options.get("lineChartType", None) == "subplots"

    def getExtraFields(self):
        if not self.isSubplot() and len(self.getValueFields())>1:
            #no clusterby if we are grouped and multiValueFields
            return []
    
        clusterby = self.options.get("clusterby")
        return [clusterby] if clusterby is not None else []

    def matplotlibRender(self, fig, ax):
        subplots = self.isSubplot()
        keyFields = self.getKeyFields()
        valueFields = self.getValueFields()

        clusterby = self.options.get("clusterby")
        if clusterby is not None and (subplots or len(valueFields)<=1):
            subplots = subplots if len(valueFields)==1 else False
            for j, valueField in enumerate(valueFields):
                pivot = self.getWorkingPandasDataFrame().pivot(
                    index=keyFields[0], columns=clusterby, values=valueField
                )
                pivot.index.name=keyFields[0]
                thisAx = pivot.plot(kind='line', ax=self.getAxItem(ax, j), sharex=True, legend=self.showLegend(), label=valueField, 
                    subplots=subplots,colormap = Colors.colormap,
                    logx=self.getBooleanOption("logx", False), logy=self.getBooleanOption("logy",False))

                if len(valueFields)==1 and subplots:
                    if isinstance(thisAx, (list,np.ndarray)):
                        #resize the figure
                        figw = fig.get_size_inches()[0]
                        fig.set_size_inches( figw, (figw * 0.5)*min( len(thisAx), 10 ))
                    return thisAx
        else:
            self.getWorkingPandasDataFrame().plot(
                kind='line', x=keyFields, y=valueFields, ax=ax, subplots=subplots, legend=self.showLegend(), colormap = Colors.colormap,
                logx=self.getBooleanOption("logx", False), logy=self.getBooleanOption("logy",False)
            )

            if clusterby is not None:
                self.addMessage("Warning: 'Cluster By' ignored when grouped option with multiple Value Fields is selected")
        
        if self.useMpld3:
            #import mpld3
            #tooltip = mpld3.plugins.PointLabelTooltip(lines[0], labels=ys)
            #mpld3.plugins.connect(fig, tooltip)
            #FIXME
            pass


# -------------------------------------------------------------------------------
# Copyright IBM Corp. 2017
# 
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
# http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -------------------------------------------------------------------------------

from pixiedust.display.chart.renderers import PixiedustRenderer
from ..baseChartDisplay import BaseChartDisplay
from six import with_metaclass
from abc import ABCMeta

import pixiedust

myLogger = pixiedust.getLogger(__name__)

@PixiedustRenderer(rendererId="google")
class GoogleBaseDisplay(with_metaclass(ABCMeta, BaseChartDisplay)):
    pass

    def getChartOptions(self):
        return [
            { 'name': 'mapDisplayMode',
              'description': 'Display Mode',
			  'metadata': {
					'type': "dropdown",
					'values': ["region", "markers", "text"],
					'default': "region"
				}
            },
            { 'name': 'mapRegion',
              'description': 'Region',
			  'metadata': {
					'type': "dropdown",
					'values': ["world", "US"],
					'default': "world"
              }
            }
        ]
  


# -------------------------------------------------------------------------------
# Copyright IBM Corp. 2016
# 
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
# http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -------------------------------------------------------------------------------
import numpy as np
import pandas as pd
import re
from pyspark.sql.types import *
import pixiedust.utils.dataFrameMisc as dataFrameMisc

def createDataframeAdapter(entity):
    if dataFrameMisc.isPandasDataFrame(entity):
        return PandasDataFrameAdapter(entity)
    elif dataFrameMisc.isPySparkDataFrame(entity):
        return entity
    raise ValueError("Invalid argument")

"""
Adapter interface to Spark APIs. Passed to pixiedust visualizations that expect a Spark DataFrame so they can work
with pandas dataframe with no code change.

This is Experimental, currently support only a subset of the Spark DataFrame APIs.
"""
class PandasDataFrameAdapter(object):
    def __init__(self, entity):
        self.entity = entity
        self.sparkDF = dataFrameMisc.isPySparkDataFrame(entity);

    def __getattr__(self, name):
        if self.sparkDF and hasattr(self.entity, name):
            return self.entity.__getattribute__(name)
        if name=="schema":
            return type("AdapterSchema",(),{"fields": self.getFields()})()
        elif name=="groupBy":
            return lambda cols: AdapterGroupBy(self.entity.groupby(cols))
        elif name=="dropna":
            return lambda: PandasDataFrameAdapter(pd.DataFrame(self.entity.dropna()))
        elif name=="sort":
            return lambda arg: self
        elif name=="select":
            return lambda name: PandasDataFrameAdapter(self.entity[name].reset_index())
        elif name=="orderBy":
            return lambda col: PandasDataFrameAdapter(self.entity.sort("agg",ascending=False))

        raise AttributeError("{0} attribute not found".format(name))

    def count(self):
        if self.sparkDF:
            return self.entity.count()
        else:
            return len(self.entity.index)

    def take(self,num):
        if self.sparkDF:
            return self.entity.take(num)
        else:
            df = self.entity.head(num)
            colNames = self.entity.columns.values.tolist()
            def makeJsonRow(row):
                ret = {}
                for i,v in enumerate(colNames):
                    ret[v]=row[i]
                return ret
            return [makeJsonRow(self.entity.iloc[i].values.tolist()) for i in range(0,len(df.index))]

    def getFields(self):
        if self.sparkDF:
            return self.entity.schema.fields
        else:
            #must be a pandas dataframe
            def createObj(a,b):
                return type("",(),{
                    "jsonValue":lambda self: {"type": b, "name": a}, "name":a,
                    "dataType": IntegerType() if np.issubdtype(b, np.integer) or np.issubdtype(b, np.float) else StringType()
                })()
            return [createObj(a,b) for a,b in zip(self.entity.columns, self.entity.dtypes)]

    def getTypeName(self):
        if self.sparkDF:
            return self.entity.schema.typeName()
        else:
            return "Pandas DataFrame Row"

    def toPandas(self):
        if self.sparkDF:
            return self.entity.toPandas()
        else:
            return self.entity

class AdapterGroupBy(object):
    def __init__(self, group):
        self.group = group

    def count(self):
        return PandasDataFrameAdapter(self.group.size().reset_index(name="count"))

    def agg(self,exp):
        m=re.search("(\w+?)\((.+?)\)(?:.+?(?:as\s+(\w*))|$)",str(exp),re.IGNORECASE)
        if m is None:
            raise AttributeError("call to agg with not supported expression: {0}".format(str(exp)))
        
        funcName=m.group(1).upper()
        groupedCol=m.group(2)
        alias=m.group(3) or "agg"
        if funcName=="SUM":
            return PandasDataFrameAdapter(self.group[groupedCol].sum().reset_index(name=alias))
        elif funcName=="AVG":
            return PandasDataFrameAdapter(self.group[groupedCol].mean().reset_index(name=alias))
        elif funcName == "MIN":
            return PandasDataFrameAdapter(self.group[groupedCol].min().reset_index(name=alias))
        elif funcName == "MAX":
            return PandasDataFrameAdapter(self.group[groupedCol].max().reset_index(name=alias))
        elif funcName == "COUNT":
            return PandasDataFrameAdapter(self.group[groupedCol].count().reset_index(name=alias))
        else:
            raise AttributeError("Unsupported aggregation function {0}".format(funcName))